/* module.c - Managing modules

   Copyright (c) 2010-2011 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include "module.h"
#include "parse.h"
#include "class.h"
#include "ident.h"
#include "memberid.h"
#include "mem.h"
#include "gc.h"
#include "dynaload.h"
#include "internal.h"
#include "std_module.h"
#include "debug_runtime.h"


#define INITIAL_CMODULES_SIZE 32


static AModuleDef **CModules;
static int CModulesSize;
static int NumCModules;


static ABool AddCModule(AModuleDef *mod, int *num);
static ABool CreateCFunction(AModuleDef *fnDef, int *num);
static ABool CreateCMethod(ASymbolInfo *class_, AModuleDef *fnDef,
                           unsigned *num);
static ABool CreateCFunctionObject(unsigned num, ASymbolInfo *sym,
                                   AModuleDef *fnDef);
static AModuleDef *CreateCType(AModuleDef *clDef);
static ASymbolInfo *GetGlobalVariable(const char *name);
static ABool AddMethod(ASymbolInfo *class_, AModuleDef *method,
                       AMemberTableType tableType, unsigned flag);
static ABool AddMemberVar(ATypeInfo *type, const char *name, ABool isWritable,
                          ABool isPrivate, int *num);
static ABool ImportModuleUsingName(const char *name);


#define ModDefName(def) (*(def)->str == '-' ? (def)->str + 1 : (def)->str)
#define IsPrivateModDef(def) (*(def)->str == '-')


ABool AInitializeCModules(void)
{
    CModulesSize = INITIAL_CMODULES_SIZE;
    
    CModules = AAllocStatic(sizeof(AModuleDef *) * CModulesSize);
    if (CModules == NULL)
        return FALSE;

    NumCModules = 0;

    return TRUE;
}


static AToken *ParseModuleName(AToken *tok, AModuleId *mod)
{
    mod->numParts = 0;
    
    /* Parse components of module name separated by quotes. */
    for (;;) {
        mod->id[mod->numParts] = (ASymbolInfo *)tok->info.sym;
        mod->numParts++;
        tok = AAdvanceTok(tok);
        if (tok->type != TT_SCOPEOP || (tok + 1)->type != TT_ID)
            break;
        tok = AAdvanceTok(tok);
    }

    return tok;
}


/* Evaluate whether the module whose identifier starts at tok has been
   compiled. Store the module id symbols at *mod. Create empty module id
   symbols for some module name parts as needed. */
AToken *AFindModule(AToken *tok, AModuleId *mod, ABool *isCompiled)
{
    /* IDEA: Document the interface and refactor a bit. */
    int i;

    /* tok->type is assumed to be TT_ID. */

    tok = ParseModuleName(tok, mod);

    i = 0;

    /* Find out whether some or all of the module id parts have already
       been generated. */
    {
        ASymbolInfo *sym = NULL;

        /* Find the last existing part of module id. */
        for (i = 0; i < mod->numParts; i++) {
            /* Is there a module symbol whose module == sym? */
            ASymbolInfo *nextSym = mod->id[i]->next;
            while (AIsModuleId(nextSym->type)
                   || nextSym->type == ID_MEMBER) {
                if (AIsModuleId(nextSym->type) && nextSym->sym == sym) {
                    mod->id[i] = nextSym;
                    goto FoundPart;
                }
                nextSym = nextSym->next;
            }

            /* Exit the loop because there wasn't a match. */
            break;

          FoundPart:

            sym = nextSym;
        }

        /* Now i is the number of found parts. In particular, if
           i == mod.numParts, every part of the module id could be found and
           the module exists unless the last part is empty. */
    }

    /* Do we have to compile the module? */
    if (i != mod->numParts
        || mod->id[i - 1]->type == ID_GLOBAL_MODULE_EMPTY
        || mod->id[i - 1]->type == ID_GLOBAL_MODULE_SUB_EMPTY) {
        /* The module hasn't been compiled yet. i points to the point
           at which we have to start generating new module symbols. */

        for (; i < mod->numParts; i++) {
            /* Generate an empty module. */
            ASymbolInfo *sym = AAddIdentifier((ASymbol *)mod->id[i],
                                            (i == 0) ? ID_GLOBAL_MODULE
                                                     : ID_GLOBAL_MODULE_SUB);
            if (sym != NULL) {
                sym->sym  = (i == 0) ? NULL : mod->id[i - 1];
                sym->type = (i == 0) ? ID_GLOBAL_MODULE_EMPTY
                                     : ID_GLOBAL_MODULE_SUB_EMPTY;
                sym->info.module.isActive   = FALSE;
                sym->info.module.isImported = FALSE;
                sym->info.module.cModule    = A_CM_NON_C;
                mod->id[i] = sym;
            } else
                return NULL;
        }

        /* Mark the last part of module id as compiled. */
        mod->id[i - 1]->type = (i == 1) ? ID_GLOBAL_MODULE
                                        : ID_GLOBAL_MODULE_SUB;

        *isCompiled = FALSE;
    } else
        *isCompiled =
            mod->id[mod->numParts - 1]->info.module.cModule == A_CM_NON_C;

    return tok;
}


/* Create a C module. */
ABool ACreateModule(AModuleDef *module, ABool existsAlready)
{
    AToken *tok;
    AToken *tokTmp;
    ABool isCompiled;
    AModuleId modId;
    char tokName[A_MAX_MODULE_NAME_LENGTH + 2];
    int num;

    if (module->type != MD_MODULE_NAME)
        return FALSE;

    strcpy(tokName, module->str);
    strcat(tokName, "\n");
    
    module++;

    tok = NULL;
    if (!ATokenize((unsigned char *)tokName,
                   (unsigned char *)tokName + strlen(tokName), &tok, &tokTmp,
                   NULL))
        return FALSE;

    if (!AddCModule(module, &num))
        return FALSE;

    /* Create the symbols for the module name parts. */
    if (!AFindModule(tok, &modId, &isCompiled)
        || (!existsAlready && isCompiled)) {
        NumCModules--;
        return FALSE;
    }

    /* Mark the module as a C module with a specific index. */
    modId.id[modId.numParts - 1]->info.module.cModule = num | 1024;
    
    return TRUE;
}


/* Create the contents of a C module (the symbol table + global variables).
   NOTE: Compilation must be active. */
ABool ARealizeModule(ASymbolInfo *module)
{
    AModuleDef *modDef;
    int num;
    int firstVar;
    int firstConst;

    modDef = CModules[module->info.module.cModule & ~1024];

    /* Realize/compile any other modules needed by the current module. */
    for (; modDef->type == MD_IMPORT; modDef++) {
        if (!ImportModuleUsingName(modDef->str))
            return FALSE;
    }

    if (AIsDynamicCompile && !AAllocateModuleGlobals(&firstVar, &firstConst))
        return FALSE;

    ABeginDynamicModuleCompilation();
    
    ACurModule = module;
    
    /* FIX: memory leak if fail? EndDynamicModuleCompilation()? (no help?) */

    for (; modDef->type != MD_END_MODULE; modDef++) {
        switch (modDef->type) {
        case MD_DEF:
            if (!CreateCFunction(modDef, &num)) {
                /* FIX: Free stuff! */
                return FALSE;
            }
            
            /* FIX: Handle error? */
            if (strcmp(modDef->str, "Main") == 0
                || strcmp(modDef->str, "-Main") == 0)
                AInitFunctions = AAddIntListEnd(AInitFunctions, num);
            
            if (modDef->numPtr != NULL)
                *modDef->numPtr = num;
            break;
            
        case MD_VAR:
        case MD_CONST:
        case MD_EMPTY_CONST: {
            AIdType type;
            ASymbol *sym;
            ASymbolInfo *symInfo;
            const char *str = ModDefName(modDef);
            ABool isPrivate = IsPrivateModDef(modDef);

            if (!AGetSymbol(str, strlen(str), &sym)) {
                /* FIX? */
                return FALSE;
            }
            
            type = modDef->type == MD_VAR ? ID_GLOBAL : ID_GLOBAL_CONST;
            symInfo = AAddGlobalVariable(sym, module, type, isPrivate);
            if (symInfo == NULL) {
                /* COSH! FIX: do something? */
                return FALSE;
            }
            
            /* FIX: check return value of both CreateConstant and SetConst.. */
            if (modDef->type == MD_CONST)
                ASetConstGlobalValue(ACompilerThread, symInfo->num,
                                    ACreateConstant(symInfo));
            
            if (modDef->numPtr != NULL)
                *modDef->numPtr = symInfo->num;
            break;
        }
            
        case MD_CLASS:
        case MD_INTERFACE:
            modDef = CreateCType(modDef);
            if (modDef == NULL) {
                /* FIX: Free memory? */
                return FALSE;
            }
            break;
        }
    }

    module->info.module.cModule = A_CM_ACTIVE;

    /* FIX: what if EndDynamicModuleCompilation fails? */

    if (AIsDynamicCompile) {
        return AEndDynamicModuleCompilation(firstVar, firstConst);
    } else
        return TRUE;
}


#define MinArgs(def) ((def)->num1)
#define MaxArgs(def) ((def)->num2)
#define ANumLocals(def) ((def)->num3)


/* FIX: what if something fails? what are we going to do? */
static ABool CreateCFunction(AModuleDef *fnDef, int *num)
{
    ASymbol *sym;
    ASymbolInfo *symInfo;
    const char *str;
    ABool isPrivate;

    str = ModDefName(fnDef);
    isPrivate = IsPrivateModDef(fnDef);

    /* The Main function is always private. */
    if (strcmp(str, "Main") == 0)
        isPrivate = TRUE;

    if (!AGetSymbol(str, strlen(str), &sym))
        return FALSE;

    symInfo = AAddGlobalVariable(sym, ACurModule, ID_GLOBAL_DEF, isPrivate);
    if (symInfo == NULL)
        return FALSE;

    if (num != NULL)
        *num = symInfo->num;
    
    symInfo->info.global.minArgs = MinArgs(fnDef);
    symInfo->info.global.maxArgs = MaxArgs(fnDef);

    return CreateCFunctionObject(symInfo->num, symInfo, fnDef);
}


static ABool CreateCMethod(ASymbolInfo *class_, AModuleDef *fnDef,
                           unsigned *num)
{
    *num = AAddConstGlobalValue(AZero);
    return CreateCFunctionObject(*num, class_, fnDef);
}


static ABool CreateCFunctionObject(unsigned num, ASymbolInfo *sym,
                                   AModuleDef *fnDef)
{
    AFunction *func;
    
    func = AAllocUnmovable(sizeof(AFunction));
    if (func == NULL)
        return FALSE;

    AInitNonPointerBlockOld(&func->header, sizeof(AFunction) -
                           sizeof(AValue) - 1);

    func->sym = sym;
    func->minArgs = MinArgs(fnDef);
    func->maxArgs = MaxArgs(fnDef);
    func->stackFrameSize = (3 + ANumLocals(fnDef)) * sizeof(AValue);
    func->codeLen = 0; /* Dummy initializer. */
#ifdef HAVE_JIT_COMPILER
    func->num = num;
#endif
    func->code.cfunc = fnDef->func;

    return ASetConstGlobalValue(ACompilerThread, num, AFunctionToValue(func));
}


static AModuleDef EmptyCreateDef = {
    MD_METHOD, 1, "-create", 1, 1, AEmptyCreate, NULL
};


/* FIX: what about errors? */
static AModuleDef *CreateCType(AModuleDef *clDef)
{
    ASymbol *sym;
    ASymbolInfo *symInfo;
    ATypeInfo *type;
    ASymbolInfo *super;
    const char *str;
    ABool isPrivate;
    ABool isInterface;
    AModuleDef *create;
    AModuleDef *def;
    AModuleDef *cur;
    int num;

    ADebugCompilerMsg(("Start creating C type"));
    ADebugVerifyMemory();

    str = ModDefName(clDef);
    isPrivate = IsPrivateModDef(clDef);
    isInterface = clDef->type == MD_INTERFACE;
    
    if (!AGetSymbol(str, strlen(str), &sym))
        return NULL;

    symInfo = AAddGlobalVariable(sym, ACurModule,
                                 isInterface ? ID_GLOBAL_INTERFACE :
                                 ID_GLOBAL_CLASS, isPrivate);
    if (symInfo == NULL)
        return NULL;
    
    type = ACreateTypeInfo(ACompilerThread, symInfo->num);
    if (type == NULL)
        return NULL;
    type->sym = symInfo;
    type->numVars = clDef->num1;
    type->isInterface = isInterface;

    /* Find supertype. */
    def = clDef + 1;
    if (clDef->numPtr == &AStdObjectNum)
        type->super = NULL; /* std::Object has no superclass. */
    else {
        if (def->type != MD_INHERIT && !isInterface) {
            /* Default to inheriting from Object. */
            super = GetGlobalVariable("std::Object");
            if (super == NULL)
                return NULL;
        } else if (def->type != MD_INHERIT && isInterface)
            super = NULL;
        else {
            super = GetGlobalVariable(def->str);
            if (super == NULL)
                return NULL;
            def++;
        }

        if (super != NULL)
            type->super = AValueToType(AGlobalByNum(super->num));
        else
            type->super = NULL;
    }

    /* Find implemented interfaces. */
    if (def->type == MD_IMPLEMENT) {
        int i;
        int numInterfaces = 0;
        do {
            numInterfaces++;
        } while (def[numInterfaces].type == MD_IMPLEMENT);
        if (!AAllocateInterfacesInTypeInfo(type, numInterfaces))
            return NULL;
        for (i = 0; i < numInterfaces; i++) {
            ASymbolInfo *sym = GetGlobalVariable(def[i].str);
            ATypeInfo *iface;
            if (sym == NULL)
                return NULL;
            iface = AValueToType(AGlobalByNum(sym->num));
            AAddInterface(type, iface, FALSE);
        }
        def += numInterfaces;
    }

    create = NULL;
    if (!isInterface) {
        /* Find constructor. */
        create = clDef + 1;
        while (create->type != MD_END_TYPE &&
               (create->type != MD_METHOD ||
                 (strcmp(create->str, "create") != 0 &&
                  strcmp(create->str, "-create") != 0)))
            create++;

        if (create->type == MD_END_TYPE) 
            create = type->super == NULL ? &EmptyCreateDef : NULL;

        if (create != NULL) {
            symInfo->info.global.minArgs = create->num1 - 1;
            symInfo->info.global.maxArgs = create->num2 - 1;
        } else {
            /* Inherit create. */
            symInfo->info.global.minArgs =
                type->super->sym->info.global.minArgs;
            symInfo->info.global.maxArgs =
                type->super->sym->info.global.maxArgs;
        }

        /* Check if there is a #f method. If it is present, allocate member
           slot 0 for the finalize object list and setup some flags. */
        for (cur = clDef + 1; cur->type != MD_END_TYPE; cur++) {
            if (cur->type == MD_METHOD && strcmp(cur->str, "#f") == 0) {
                /* Found it. Allocate a slot. Logically, the slot 0 is
                   allocated for the list, but we do not need to care about it
                   at this point. */
                type->hasFinalizer = TRUE;
                type->hasFinalizerOrData = TRUE;
                type->numVars++;
                break;
            }
        }
    }

    for (; def->type != MD_END_TYPE; def++) {
        const char *str = ModDefName(def);
        
        if (IsPrivateModDef(def) && def->type != MD_VAR) {
            /* FIX: Free stuff? */
            return NULL;
        }
        
        switch (def->type) {
        case MD_METHOD:
            if (strcmp(def->str, "create") != 0) {
                if (!AddMethod(symInfo, def, MT_METHOD_PUBLIC, 0))
                    return NULL;
            }
            break;
            
        case MD_GETTER:
            if (!AddMethod(symInfo, def, MT_VAR_GET_PUBLIC, A_VAR_METHOD))
                return NULL;
            break;
            
        case MD_SETTER:
            if (!AddMethod(symInfo, def, MT_VAR_SET_PUBLIC, A_VAR_METHOD))
                return NULL;
            break;
            
        case MD_VAR:
            if (!AddMemberVar(type, str, TRUE, IsPrivateModDef(def), &num))
                return NULL;
            if (def->numPtr != NULL)
                *def->numPtr = num + ATotalNumMemberVars(type->super);
            break;
            
        case MD_EMPTY_CONST:
            if (!AddMemberVar(type, def->str, FALSE, FALSE, &num))
                return NULL;
            if (def->numPtr != NULL)
                *def->numPtr = num + ATotalNumMemberVars(type->super);
            break;

        case MD_BINARY_DATA:
            break;

        case MD_EXTERNAL_DATA:
            if (!AddMemberVar(type, "#ext", TRUE, TRUE, &num))
                return NULL;
            type->extDataMember = num + ATotalNumMemberVars(type->super);
            break;

        default:
            return NULL;
        }
    }

    if (!isInterface) {
        if (create != NULL) {
            if (!CreateCMethod(symInfo, create, &type->create))
                return FALSE;
            AAddMember(MT_METHOD_PUBLIC, AM_CREATE, type->create);
        } else {
            /* Inherit create. */
            type->create = type->super->create;
        }
    }

    ABuildTypeInfoMembers(type); /* FIX: what if out of memory? */
    AUpdateInheritedMiscTypeData(type);
    
    AUpdateTypeTotalNumVars(type);
    type->totalNumVars += type->numVars;

    /* Figure binary data offsets. */
    type->dataSize = 0;
    for (def = clDef; def->type != MD_END_TYPE ; def++) {
        if (def->type == MD_BINARY_DATA) {
            if (def->numPtr != NULL) {
                int offset = type->super != NULL ? type->super->dataSize : 0;
                *def->numPtr = offset + type->dataSize;
            }
            type->dataSize += def->num1;
            break;
        }
    }

    ACalculateInstanceSize(type);

    if (clDef->numPtr != NULL)
        *clDef->numPtr = symInfo->num;

    ADebugVerifyMemory();
    ADebugCompilerMsg(("Created C class"));

    return def;
}


static ABool AddMethod(ASymbolInfo *class_, AModuleDef *method,
                       AMemberTableType tableType, unsigned flag)
{
    /* FIX: error checking sucks */
    
    ASymbol *sym;
    ASymbolInfo *symInfo;
    unsigned memberNum;
    unsigned globalNum;
    
    if (method->str[0] != '#') {
        if (!AGetSymbol(method->str, strlen(method->str),
                       &sym))
            return FALSE;
        
        symInfo = AGetMemberSymbol(sym);
        if (symInfo == NULL)
            return FALSE;
        
        memberNum = symInfo->num;
    } else {
        if (method->str[1] == 'i')
            memberNum = AM_INITIALIZER;
        else
            memberNum = AM_FINALIZER;
    }
    
    if (!CreateCMethod(class_, method, &globalNum))
        return FALSE;
    
    AAddMember(tableType, memberNum, globalNum | flag);

    return TRUE;
}


static ABool AddMemberVar(ATypeInfo *type, const char *name, ABool isWritable,
                          ABool isPrivate, int *num)
{
    /* FIX: out of memory? */
    ASymbol *sym;
    ASymbolInfo *symInfo;
    unsigned key;
    unsigned item;
    
    if (!AGetSymbol(name, strlen(name), &sym))
        return FALSE;
    
    symInfo = AGetMemberSymbol(sym);
    if (symInfo == NULL)
        return FALSE;
    
    key = symInfo->num;
    *num = type->numVars;
    item = type->numVars++;
    
    AAddMember(isPrivate ? MT_VAR_GET_PRIVATE : MT_VAR_GET_PUBLIC, key, item);
    if (isWritable)
        AAddMember(isPrivate ? MT_VAR_SET_PRIVATE : MT_VAR_SET_PUBLIC, key,
                  item);

    return TRUE;
}


/* Store a C module in the global module list. Store the index of the module
   in *num. Return FALSE if out of memory. */
static ABool AddCModule(AModuleDef *mod, int *num)
{
    if (NumCModules == CModulesSize) {
        CModulesSize *= 2;
        CModules = AGrowStatic(CModules, sizeof(AModuleDef *) * CModulesSize);
        if (CModules == NULL)
            return FALSE;
    }

    *num = NumCModules;
    CModules[NumCModules] = mod;
    
    NumCModules++;

    return TRUE;
}


AValue AEmptyCreate(AThread *t, AValue *args)
{
    return args[0];
}


AValue AEmptyMain(AThread *t, AValue *args)
{
    return ANil;
}


/* NOTE: The name must be valid! Otherwise, unexpected things may happen. */
static ASymbolInfo *GetGlobalVariable(const char *name)
{
    AToken *tok;
    AToken *tokTmp;
    char tokName[A_MAX_MODULE_NAME_LENGTH + 2];
    ASymbolInfo *sym;    

    strcpy(tokName, name);
    strcat(tokName, "\n");

    tok = NULL;
    if (!ATokenize((unsigned char *)tokName,
                   (unsigned char *)tokName + strlen(tokName), &tok, &tokTmp,
                   NULL))
        return NULL;

    AParseGlobalVariable(tok, &sym, TRUE, TRUE);
    /* FIX: Free tokens! */
    
    /* IDEA: The following test is fragile. */
    if (sym->type == ID_ERR_PARSE ||
        (sym->type == ID_GLOBAL && sym->num == 0))
        return NULL;
    else
        return sym;
}


ABool AInitializeModule(char *name, ABool isAutoImport)
{
    AToken *tok;
    AToken *tokTmp;
    ABool isCompiled;
    AModuleId modId;
    char tokName[A_MAX_MODULE_NAME_LENGTH + 2];
    ASymbolInfo *sym;

    strcpy(tokName, name);
    strcat(tokName, "\n");

    tok = NULL;
    if (!ATokenize((unsigned char *)tokName,
                   (unsigned char *)tokName + strlen(tokName), &tok, &tokTmp,
                   NULL))
        return FALSE;

    if (!AFindModule(tok, &modId, &isCompiled) || isCompiled)
        return FALSE;

    sym = modId.id[modId.numParts - 1];
    if (!ARealizeModule(sym))
        return FALSE;

    if (isAutoImport) {
        sym->info.module.cModule = A_CM_AUTO_IMPORT;
        sym->info.module.isImported = TRUE;
        do {
            sym->info.module.isActive = TRUE;
            sym = sym->sym;
        } while (sym != NULL);
    }
    
    return TRUE;
}


static ABool ImportModuleUsingName(const char *name)
{
    char tokName[A_MAX_MODULE_NAME_LENGTH + 2];
    AToken *tok;
    AToken *tokTmp;

    /* FIX: create a function for the following */
    strcpy(tokName, name);
    strcat(tokName, "\n");

    tok = NULL;
    if (!ATokenize((unsigned char *)tokName,
                   (unsigned char *)tokName + strlen(tokName), &tok, &tokTmp,
                   NULL))
        return FALSE;

    return AImportModule(tok, TRUE) != NULL;
}


AToken *AImportModule(AToken *tok, ABool isFromC)
{
    AModuleId impMod;
    ABool isCompiled;
    ASymbolInfo *sym;
    
    tok = AFindModule(tok, &impMod, &isCompiled);
    if (tok == NULL)
        return NULL;
    
    sym = impMod.id[impMod.numParts - 1];
    
    if (isCompiled)
        goto SkipUntilNewlineOrComma;
    
    if (!isFromC) {
        /* Record the line number of the import statement so that
           it is available to be displayed in error messages. */
        ALineNums[ANumActiveFiles - 1] = tok->lineNumber;
    }
    
    if (sym->info.module.cModule == A_CM_NON_C) {
        /* Try to compile the module. */
        if (!ACompileModule(&impMod))
            return NULL;
    } else if (sym->info.module.cModule >= 0) {
        /* Non-realized C module must be realized. */
        if (!ARealizeModule(sym))
            return NULL;
    }
    
    if (tok->type != TT_NEWLINE && tok->type != TT_COMMA)
        goto SkipUntilNewlineOrComma;

    return tok;

  SkipUntilNewlineOrComma:

    while (tok->type != TT_NEWLINE && tok->type != TT_COMMA)
        tok = AAdvanceTok(tok);

    return tok;
}
