/* os_module.c - os module

   Copyright (c) 2010-2011 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

/* Implementation of the os module. Contains mostly miscellaneous file system
   related services. */

#include "aconfig.h"

#ifdef A_HAVE_OS_MODULE

#include <time.h>
#include <stdlib.h>
#include "alore.h"
#include "errmsg.h"
#include "util.h"
#include "os_module.h"
#include "internal.h"


/* IDEA: Add a function for extracting the drive portion of a Windows path? */


static int DirSeparatorNum;
static int PathSeparatorNum;


static const AMixedCValueType ArgsStr[] = {
    A_MVT_STR, A_MVT_LAST_ARG
};

static const AMixedCValueType ArgsStrOptStr[] = {
    A_MVT_STR, A_MVT_OPT_STR, A_MVT_LAST_ARG
};

static const AMixedCValueType ArgsStrStr[] = {
    A_MVT_STR, A_MVT_STR, A_MVT_LAST_ARG
};

static const AMixedCValueType ArgsStrStrOptStr[] = {
    A_MVT_STR, A_MVT_STR, A_MVT_OPT_STR, A_MVT_LAST_ARG
};

static const AMixedCValueType ArgsFloat[] = {
    A_MVT_FLOAT, A_MVT_LAST_ARG
};

static const AMixedCValueType ArgsVoid[] = {
    A_MVT_LAST_ARG
};


/* Remove a file, a directory or another name in the file system. */
static AValue Remove(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, ARemoveImpl);
}


/* Rename a file or a directory. */
static AValue Rename(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStrStr, ARenameImpl);
}


/* Create a directory. */
static AValue MakeDir(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AMakeDirImpl);
}


/* Change the current working directory. */
static AValue ChangeDir(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AChangeDirImpl);
}


/* Return the current working directory. */
static AValue CurrentDir(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsVoid, ACurrentDirImpl);
}


/* Suspend the execution of the current t for a certain period time. */
static AValue Sleep(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsFloat, ASleepImpl);
}


/* Execute an executable or a command using the command interpreter. */
static AValue System(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, ASystemImpl);
}


static AValue DirName(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];
    int i;

    i = AGetStr(t, frame[0], path, A_MAX_PATH_LEN);
    for (; i > 0 && !A_IS_DIR_SEPARATOR(path[i - 1])
             && !A_IS_DRIVE_SEPARATOR(path[i - 1]); i--);

    if (i > 0) {
        for (; A_IS_DIR_SEPARATOR(path[i - 1]); i--) {
            if (i == 1) 
                break;
        }
    }

    if (i > 0 && A_IS_DRIVE_SEPARATOR(path[i - 1]) &&
        A_IS_DIR_SEPARATOR(path[i]))
        i++;

    return ASubStr(t, frame[0], 0, i);
}


static AValue BaseName(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];
    int len, i;

    len = AGetStr(t, frame[0], path, A_MAX_PATH_LEN);
    for (i = len; i > 0 && !A_IS_DIR_SEPARATOR(path[i - 1])
             && !A_IS_DRIVE_SEPARATOR(path[i - 1]); i--);

    return ASubStr(t, frame[0], i, len);
}


static AValue FileExt(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];
    int len, i;

    len = AGetStr(t, frame[0], path, A_MAX_PATH_LEN);
    for (i = len - 1; i >= 0; i--) {
        if (A_IS_DIR_SEPARATOR(path[i])) {
            i = len;
            break;
        } else if (path[i] == '.')
            break;
    }

    if (i < 0)
        i = len;

    return ASubStr(t, frame[0], i, len);
}


static AValue FileBase(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];
    int len, i;

    len = AGetStr(t, frame[0], path, A_MAX_PATH_LEN);
    for (i = len - 1; i >= 0; i--) {
        if (A_IS_DIR_SEPARATOR(path[i])) {
            i = len;
            break;
        } else if (path[i] == '.')
            break;
    }

    if (i < 0)
        i = len;

    return ASubStr(t, frame[0], 0, i);
}


/* Join several path components. */
static AValue OsJoinPath(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];    
    int pathInd;
    int aLen;
    int i;

    pathInd = AGetStr(t, frame[0], path, A_MAX_PATH_LEN);

    aLen = AArrayLen(frame[1]);

    for (i = 0; i < aLen; i++) {
        int len;
        AValue item = AArrayItem(frame[1], i);
        AWideChar ch1;

        AExpectStr(t, item);
        len = AStrLen(item);
        if (len == 0) {
            if (pathInd > 0) {
                if (pathInd >= A_MAX_PATH_LEN - 1)
                    return ARaiseValueErrorND(t);
                if (!A_IS_DIR_SEPARATOR(path[pathInd - 1])) {
                    path[pathInd++] = A_DIR_SEPARATOR;
                    path[pathInd] = '\0';
                }
            }
            continue;
        }

        if (pathInd > 0 && !A_IS_DIR_SEPARATOR(path[pathInd - 1])) {
            if (pathInd == A_MAX_PATH_LEN)
                return ARaiseValueErrorND(t);
            path[pathInd++] = A_DIR_SEPARATOR;
        }

        ch1 = AStrItem(item, 0);
#ifndef A_HAVE_WINDOWS        
        if (A_IS_DIR_SEPARATOR(ch1))
            pathInd = 0;
#else
        {
            AWideChar ch2 = AStrLen(item) > 1 ? AStrItem(item, 1) : ' ';
            if (A_IS_DIR_SEPARATOR(ch1) || (ch2 == ':' && AIsAlpha(ch1)))
                pathInd = 0;
        }
#endif
        pathInd += AGetStr(t, AArrayItem(frame[1], i), path + pathInd,
                           A_MAX_PATH_LEN - pathInd);
    }
    
    return AMakeStr(t, path);
}


/* Does a path point to any directory entry? */
static AValue Exists(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AExistsImpl);
}


/* Does a path point to a file? */
static AValue OsIsFile(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AIsFileImpl);
}


/* Does a path point to a directory? */
static AValue IsDir(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AIsDirImpl);
}


/* Does a path point to a link? */
static AValue IsLink(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AIsLinkImpl);
}


/* Is a path absolute? */
static AValue IsAbs(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];

    AGetStr(t, frame[0], path, A_MAX_PATH_LEN);

    return A_IS_ABS(path) ? ATrue : AFalse;
}


static AValue OSMain(AThread *t, AValue *frame)
{
    ASetConstGlobalByNum(t, DirSeparatorNum,
                         AMakeStr(t, A_DIR_SEPARATOR_STRING));
    ASetConstGlobalByNum(t, PathSeparatorNum,
                         AMakeStr(t, A_PATH_SEPARATOR_STRING));
    return ANil;
}


static AValue ListDir(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStrOptStr, AListDirImpl);
}


static AValue WalkDir(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStrOptStr, AWalkDirImpl);
}


static AValue SetEnv(AThread *t, AValue *frame)
{
    if (AIsNil(frame[1]))
        return ACallOsService(t, frame, ArgsStr, AUnsetEnvImpl);
    else
        return ACallOsService(t, frame, ArgsStrStr, ASetEnvImpl);
}


static AValue GetEnv(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStr, AGetEnvImpl);
}


static AValue SetPerm(AThread *t, AValue *frame)
{
    return ACallOsService(t, frame, ArgsStrStrOptStr, ASetPermImpl);
}


/* IDEA: Add Exec() function. */
A_MODULE(os, "os")
    A_DEF(A_PRIVATE("Main"), 0, 5, OSMain)
    A_DEF("Remove", 1, 0, Remove)
    A_DEF("Rename", 2, 0, Rename)
    A_DEF("MakeDir", 1, 0, MakeDir)
    A_DEF("ChangeDir", 1, 0, ChangeDir)
    A_DEF("CurrentDir", 0, 0, CurrentDir)
    A_DEF("Sleep", 1, 0, Sleep)
    A_DEF("System", 1, 0, System)
    A_DEF_OPT("ListDir", 1, 2, 0, ListDir)
    A_DEF_OPT("WalkDir", 1, 2, 0, WalkDir)
    A_DEF("Exists", 1, 0, Exists)
    A_DEF("IsDir", 1, 0, IsDir)
    A_DEF("IsFile", 1, 0, OsIsFile)
    A_DEF("IsLink", 1, 0, IsLink)
    A_DEF("IsAbs", 1, 0, IsAbs)
    A_DEF("DirName", 1, 0, DirName)
    A_DEF("BaseName", 1, 0, BaseName)
    A_DEF("FileExt", 1, 0, FileExt)
    A_DEF("FileBase", 1, 0, FileBase)
    A_DEF_VARARG("JoinPath", 1, 1, 0, OsJoinPath)
    A_DEF("SetEnv", 2, 0, SetEnv)
    A_DEF("GetEnv", 1, 0, GetEnv)
    A_DEF("ListEnv", 0, 1, AListEnv)
    A_DEF_OPT("SetPerm", 2, 3, 0, SetPerm)
    A_DEF("SetModificationTime", 2, 0, ASetModificationTime)
    A_DEF("SetAccessTime", 2, 0, ASetAccessTime)
    A_DEF("User", 0, 0, AOsUser)

    A_EMPTY_CONST_P("DirSeparator", &DirSeparatorNum)
    A_EMPTY_CONST_P("PathSeparator", &PathSeparatorNum)

    A_CLASS_PRIV("Stat", 3)
        A_BINARY_DATA(sizeof(time_t) * 2)
        A_METHOD("create", 1, 0, AStatCreate)
        /* NOTE: StatImpl depends on the order of these variables! */
        A_EMPTY_CONST("size")
        A_EMPTY_CONST("isFile")
        A_EMPTY_CONST("isDir")
        A_EMPTY_CONST("isLink")
        A_EMPTY_CONST("isSpecial")
        A_GETTER("modificationTime", 7, AStatModificationTime)
        A_GETTER("accessTime", 7, AStatAccessTime)
        A_GETTER("isReadable", 0, AStatIsReadable)
        A_GETTER("isWritable", 0, AStatIsWritable)
        A_GETTER("owner", 0, AStatOwner)
        A_GETTER("ownerPerm", 0, AStatOwnerPerm)
        A_GETTER("otherPerm", 0, AStatOtherPerm)
    A_END_CLASS()
A_END_MODULE()


#endif /* HAVE_OS_MODULE */
