/* os_posix.c - Unix/posix specific implementation details of the os module

   Copyright (c) 2010-2011 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include "aconfig.h"

#if defined(A_HAVE_OS_MODULE) && !defined(A_HAVE_WINDOWS)

#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <dirent.h>
#include <time.h>
#include <utime.h>
#include <pwd.h>
#include "alore.h"
#include "os_module.h"
#include "io_module.h"
#include "str.h"
#include "array.h"
#include "internal.h"


#define MKDIR_MODE 0777 /* FIX: what is the correct mode? symbolic? */


static AValue MakeDateTime(AThread *t, AValue *frame, time_t t1);


AValue ARemoveImpl(AThread *t, const AMixedCValue *frame)
{
    int status;

    AAllowBlocking();
    status = remove(frame[0].strValue);
    AEndBlocking();
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue ARenameImpl(AThread *t, const AMixedCValue *frame)
{
    int status;
    
    /* IDEA: Maybe return a special error if cannot rename because destination
       on a different file system. */

    /* IDEA: Probably create the directories beneath the destination directory
       as needed. */

    AAllowBlocking();
    status = rename(frame[0].strValue, frame[1].strValue);
    AEndBlocking();        
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue AMakeDirImpl(AThread *t, const AMixedCValue *frame)
{
    int status;
    
    /* IDEA: Create the directories beneath the destination directory as
       needed. */

    AAllowBlocking();
    status = mkdir(frame[0].strValue, MKDIR_MODE);
    AEndBlocking();
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue AChangeDirImpl(AThread *t, const AMixedCValue *frame)
{
    int status;

    AAllowBlocking();
    status = chdir(frame[0].strValue);
    AEndBlocking();        
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue ACurrentDirImpl(AThread *t, const AMixedCValue *frame)
{
    char buf[A_MAX_PATH_LEN];
    
    if (getcwd(buf, A_MAX_PATH_LEN) == NULL)
        return ARaiseErrnoIoError(t);
    else
        return AMakeStr(t, buf);
}


AValue ASleepImpl(AThread *t, const AMixedCValue *frame)
{
    AAllowBlocking();
    usleep((int)(frame[0].floatValue * 1000000)); /* FIX: Rounding?
                                                          Overflow? */
    AEndBlocking();
    return ANil;
}


AValue ASystemImpl(AThread *t, const AMixedCValue *frame)
{
    /* IDEA: Can we analyze the return code and raise an exception or
       something -- but perhaps we'd better not. */
    int error;

    AAllowBlocking();
    error = system(frame[0].strValue);
    AEndBlocking();
    
    return AMakeInt(t, error);;
}


AValue AExistsImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = stat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }
    
    return ATrue;
}


AValue AIsFileImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = stat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }
    
    return S_ISREG(buf.st_mode) ? ATrue : AFalse;
}


AValue AIsDirImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = stat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }

    return S_ISDIR(buf.st_mode) ? ATrue : AFalse;
}


AValue AIsLinkImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = lstat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }

    return S_ISLNK(buf.st_mode) ? ATrue : AFalse;
}


enum {
    STAT_PATH,
    STAT_UID,
    STAT_PERM,
    STAT_SIZE,
    STAT_IS_FILE,
    STAT_IS_DIR,
    STAT_IS_LINK,
    STAT_IS_SPECIAL
};


AValue AStatCreate(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];
    struct stat buf;
    ABool isLink = FALSE;
    int status;

    AGetStr(t, frame[1], path, A_MAX_PATH_LEN);
    
    AAllowBlocking();
    status = lstat(path, &buf);
    AEndBlocking();
    
    if (status < 0)
        return ARaiseErrnoIoError(t);

    if (S_ISLNK(buf.st_mode)) {
        isLink = TRUE;
        if (stat(path, &buf) < 0)
            return ARaiseErrnoIoError(t);
    }

    ASetMemberDirect(t, frame[0], STAT_PATH, frame[1]);
    ASetMemberDirect(t, frame[0], STAT_UID, AIntToValue(buf.st_uid));
    ASetMemberDirect(t, frame[0], STAT_PERM, AIntToValue(buf.st_mode));
    /* FIX: Max size is 512MB!! */
    ASetMemberDirect(t, frame[0], STAT_SIZE, AIntToValue(buf.st_size));
    ASetMemberDirect(t, frame[0], STAT_IS_FILE, S_ISREG(buf.st_mode)
                    ? ATrue : AFalse);
    ASetMemberDirect(t, frame[0], STAT_IS_DIR, S_ISDIR(buf.st_mode)
                    ? ATrue : AFalse);
    ASetMemberDirect(t, frame[0], STAT_IS_LINK, isLink ? ATrue : AFalse);
    ASetMemberDirect(t, frame[0], STAT_IS_SPECIAL, S_ISCHR(buf.st_mode) ||
                    S_ISBLK(buf.st_mode) || S_ISFIFO(buf.st_mode) ||
                    S_ISSOCK(buf.st_mode) ? ATrue : AFalse);

    *(time_t *)ADataPtr(frame[0], 0) = buf.st_mtime;
    *(time_t *)ADataPtr(frame[0], sizeof(time_t)) = buf.st_atime;
    
    return frame[0];
}


AValue AStatModificationTime(AThread *t, AValue *frame)
{
    time_t t1 = *(time_t *)ADataPtr(frame[0], 0);
    return MakeDateTime(t, frame, t1);
}


AValue AStatAccessTime(AThread *t, AValue *frame)
{
    time_t t1 = *(time_t *)ADataPtr(frame[0], sizeof(time_t));
    return MakeDateTime(t, frame, t1);
}


static AValue MakeDateTime(AThread *t, AValue *frame, time_t t1)
{
    struct tm t2;
    char buf[30];

    localtime_r(&t1, &t2);
    
    sprintf(buf, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d",
            t2.tm_year + 1900,
            t2.tm_mon + 1,
            t2.tm_mday,
            t2.tm_hour,
            t2.tm_min,
            AMin(59, t2.tm_sec));

    return AMakeStr(t, buf);
}


AValue AStatIsReadable(AThread *t, AValue *frame)
{
    AValue pathVal = AMemberDirect(frame[0], STAT_PATH);
    char path[A_MAX_PATH_LEN];
    int status;
    
    AGetStr(t, pathVal, path, A_MAX_PATH_LEN);

    AAllowBlocking();
    status = access(path, R_OK);
    AEndBlocking();
    
    return status == 0 ? ATrue : AFalse;
}


AValue AStatIsWritable(AThread *t, AValue *frame)
{
    AValue pathVal = AMemberDirect(frame[0], STAT_PATH);
    char path[A_MAX_PATH_LEN];
    int status;
    
    AGetStr(t, pathVal, path, A_MAX_PATH_LEN);

    AAllowBlocking();
    status = access(path, W_OK);
    AEndBlocking();
    
    return status == 0 ? ATrue : AFalse;
}


static AValue UidToStr(AThread *t, int uid)
{
    /* IDEA: Perhaps there should some way of caching these results, since the
             getpwuid_r() call might be slow. */
    
    struct passwd pwbuf;
    char buf[2048];

#ifdef HAVE_GETPWUID_R_POSIX
    int status;
    struct passwd *pwbufptr;
    status = getpwuid_r(uid, &pwbuf, buf, sizeof(buf), &pwbufptr);
    if (status != 0) {
        /* getpwuid_r returns an error number but does not modify errno
           according to Linux manpages. */
        errno = status;
        return ARaiseErrnoIoError(t);
    }
#else
    /* Solaris getpwuid_r is slightly different... */
    if (getpwuid_r(uid, &pwbuf, buf, sizeof(buf)) == NULL)
        return ARaiseErrnoIoError(t);
#endif
    
    return AMakeStr(t, pwbuf.pw_name);
}

        
AValue AStatOwner(AThread *t, AValue *frame)
{
    int uid = AValueToInt(AMemberDirect(frame[0], STAT_UID));
    return UidToStr(t, uid);
}


static AValue MakePermStr(AThread *t, int perm)
{
    char buf[4];
    int i = 0;

    if (perm & 4)
        buf[i++] = 'r';
    if (perm & 2)
        buf[i++] = 'w';
    if (perm & 1)
        buf[i++] = 'x';
    buf[i] = '\0';
    return AMakeStr(t, buf);
}


AValue AStatOwnerPerm(AThread *t, AValue *frame)
{
    int perm = AValueToInt(AMemberDirect(frame[0], STAT_PERM));
    return MakePermStr(t, (perm >> 6) & 7);
}


AValue AStatOtherPerm(AThread *t, AValue *frame)
{
    int perm = AValueToInt(AMemberDirect(frame[0], STAT_PERM));
    return MakePermStr(t, perm & 7);
}


static ABool MatchName(const char *str, const char *mask)
{
    for (; *mask != '\0'; mask++) {
        if (*mask == '*') {
            const char *s = str;
            while (*s != '\0')
                s++;
            for (; s >= str; s--) {
                if (MatchName(s, mask + 1))
                    return TRUE;
            }
            return FALSE;
        } else if (*mask == '?') {
            if (*str == '\0')
                return FALSE;
            str++;
        } else {
            if (*mask != *str)
                return FALSE;
            str++;
        }
    }

    return *str == '\0';
}


AValue AListDirImpl(AThread *t, const AMixedCValue *frame)
{
    DIR *d;
    AValue *tmp;
    AValue ret;
    struct dirent *entry;

    tmp = AAllocTemps(t, 2);

    AAllowBlocking();
    d = opendir(frame[0].strValue);
    AEndBlocking();
    
    if (d == NULL) {
        AFreeTemps(t, 2);
        return ARaiseErrnoIoError(t);
    }

    if (ATry(t)) {
        AAllowBlocking();
        closedir(d);
        AEndBlocking();
        return AError;
    }
    
    tmp[1] = AZero; /* IDEA: Is this superfluous? */
    tmp[0] = AMakeArray(t, 0);

    for (;;) {
        const char *n;

        AAllowBlocking();
        entry = readdir(d);
        AEndBlocking();
        
        if (entry == NULL)
            break;
        
        n = entry->d_name;
        
        if (!(n[0] == '.' && (strcmp(n, ".") == 0 || strcmp(n, "..") == 0))) {
            if (frame[1].strValue == NULL
                || MatchName(n, frame[1].strValue)) {
                tmp[1] = AMakeStr(t, n);
                AAppendArray(t, tmp[0], tmp[1]);
            }
        }
    }

    AEndTry(t);

    AAllowBlocking();
    closedir(d);
    AEndBlocking();

    ret = tmp[0];
    AFreeTemps(t, 2);
    return ret;
}


static AValue WalkDirRecursive(AThread *t, AValue *data, const char *base,
                               char *rel, const char *mask)
{
    DIR *d;
    struct dirent *entry;
    struct stat buf;
    char path[A_MAX_PATH_LEN];

    strcpy(path, base);
    strcat(path, A_DIR_SEPARATOR_STRING);
    strcat(path, rel);

    AAllowBlocking();
    d = opendir(path);
    AEndBlocking();
    
    if (d == NULL)
        return ARaiseErrnoIoError(t);
    
    if (ATry(t)) {
        AAllowBlocking();
        closedir(d);
        AEndBlocking();
        return AError;
    }

    for (;;) {
        const char *n;
        int status;

        AAllowBlocking();
        entry = readdir(d);
        AEndBlocking();
        
        if (entry == NULL)
            break;
        
        n = entry->d_name;
        
        if (n[0] == '.' && (strcmp(n, ".") == 0 || strcmp(n, "..") == 0))
            continue;

        strcpy(path, base);
        strcat(path, A_DIR_SEPARATOR_STRING);
        strcat(path, rel);
        strcat(path, A_DIR_SEPARATOR_STRING);
        strcat(path, n);

        AAllowBlocking();
        status = stat(path, &buf);
        AEndBlocking();
        
        if (status == -1) {
            AAllowBlocking();
            closedir(d);
            AEndBlocking();
            return ARaiseErrnoIoError(t);
        }

        /* Concatenate n to rel and store the result in path. */
        strcpy(path, rel);
        if (*path != '\0')
            strcat(path, A_DIR_SEPARATOR_STRING);
        strcat(path, n);
        
        if (mask == NULL || MatchName(n, mask)) {
            data[1] = AMakeStr(t, path);
            AAppendArray(t, data[0], data[1]);
        }
        
        /* Descent into non-link directories. */
        if (S_ISDIR(buf.st_mode) && !S_ISLNK(buf.st_mode)) {
            if (AIsError(WalkDirRecursive(t, data, base, path, mask))) {
                AAllowBlocking();
                closedir(d);
                AEndBlocking();
                return AError;
            }
        }
    }

    AEndTry(t);

    AAllowBlocking();
    closedir(d);
    AEndBlocking();
    
    return data[0];
}


AValue AWalkDirImpl(AThread *t, const AMixedCValue *frame)
{
    AValue *tmp;
    AValue ret;

    tmp = AAllocTemps(t, 2);
    tmp[1] = AZero; /* IDEA: Is this superfluous? */
    tmp[0] = AMakeArray(t, 0);

    ret = WalkDirRecursive(t, tmp, frame[0].strValue, "", frame[1].strValue);

    AFreeTemps(t, 2);
    return ret;
}


AValue AGetEnvImpl(AThread *t, const AMixedCValue *frame)
{
    char *value = getenv(frame[0].strValue);
    
    return value == NULL ? ANil : AMakeStr(t, value);
}


AValue ASetEnvImpl(AThread *t, const AMixedCValue *frame)
{
    if (setenv(frame[0].strValue, frame[1].strValue, TRUE) < 0)
        return ARaiseByNum(t, AErrorClassNum[EX_RESOURCE_ERROR], NULL);
    else
        return ANil;
}


AValue AUnsetEnvImpl(AThread *t, const AMixedCValue *frame)
{
    unsetenv(frame[0].strValue);
    return ANil;
}


static int PermBits(AThread *t, const char *str)
{
    int bits = 0;
    for (; *str != '\0'; str++) {
        if (*str == 'r')
            bits |= 4;
        else if (*str == 'w')
            bits |= 2;
        else if (*str == 'x')
            bits |= 1;
        else
            ARaiseValueError(t, NULL);
    }
    return bits;
}


AValue ASetPermImpl(AThread *t, const AMixedCValue *frame)
{
    int userPerm, groupPerm, otherPerm;
    int mode;
    int status;

    userPerm = PermBits(t, frame[1].strValue);
    if (frame[2].strValue != NULL)
        otherPerm = PermBits(t, frame[2].strValue);
    else
        otherPerm = userPerm & ~2; /* Copy all but bits but the write bit. */
    groupPerm = otherPerm;
    
    mode = (userPerm << 6) | (groupPerm << 3) | otherPerm;

    AAllowBlocking();
    status = chmod(frame[0].strValue, mode);
    AEndBlocking();
    
    if (status == -1)
        return ARaiseErrnoIoError(t);

    return ANil;
}


static AValue SetFileTime(AThread *t, AValue *frame, ABool isMod)
{
    char path[A_MAX_PATH_LEN];
    AValue v;
    struct tm tm;
    time_t tim;
    struct utimbuf utim;
    struct stat statbuf;
    int status;

    AGetStr(t, frame[0], path, sizeof(path));

    v = AMember(t, frame[1], "year");
    if (AIsError(v))
        return AError;
    tm.tm_year = AGetInt(t, v) - 1900;

    v = AMember(t, frame[1], "month");
    if (AIsError(v))
        return AError;
    tm.tm_mon = AGetInt(t, v) - 1;

    v = AMember(t, frame[1], "day");
    if (AIsError(v))
        return AError;
    tm.tm_mday = AGetInt(t, v);

    v = AMember(t, frame[1], "hour");
    if (AIsError(v))
        return AError;
    tm.tm_hour = AGetInt(t, v);

    v = AMember(t, frame[1], "minute");
    if (AIsError(v))
        return AError;
    tm.tm_min = AGetInt(t, v);

    v = AMember(t, frame[1], "second");
    if (AIsError(v))
        return AError;
    tm.tm_sec = AGetFloat(t, v);

    tm.tm_isdst = -1;

    tim = mktime(&tm);
    if (tim == (time_t)-1)
        return ARaiseValueError(t, NULL); /* IDEA: Message */

    /* Read the current access and modification times using stat. This needed
       since utime() changes both of them -- we must set the unchanged value
       to its current value. */
    AAllowBlocking();
    status = stat(path, &statbuf);
    AEndBlocking();
    if (status == -1)
        return ARaiseErrnoIoError(t);

    if (isMod) {
        utim.actime = statbuf.st_atime;
        utim.modtime = tim;
    } else {
        utim.actime = tim;
        utim.modtime = statbuf.st_mtime;
    }

    AAllowBlocking();
    status = utime(path, &utim);
    AEndBlocking();    
    if (status == -1)
        return ARaiseErrnoIoError(t);

    return ANil;
}


AValue ASetModificationTime(AThread *t, AValue *frame)
{
    return SetFileTime(t, frame, TRUE);
}


AValue ASetAccessTime(AThread *t, AValue *frame)
{
    return SetFileTime(t, frame, FALSE);
}


extern char **environ;


AValue AListEnv(AThread *t, AValue *frame)
{
    int len;
    int i;

    for (len = 0; environ[len] != NULL; len++);

    frame[0] = AMakeArray(t, len);

    for (i = 0; i < len; i++) {
        char *s = environ[i];
        char var[1024];
        int j;
        AValue ss;

        j = 0;
        for (j = 0; j < 1023 && s[j] != '\0' && s[j] != '='; j++)
            var[j] = s[j];
        var[j] = '\0';

        ss = AMakeStr(t, var);
        ASetArrayItem(t, frame[0], i, ss);
    }

    return frame[0];
}


AValue AOsUser(AThread *t, AValue *frame)
{
    return UidToStr(t, getuid());
}


#endif /* defined(A_HAVE_OS_MODULE) && !defined(A_HAVE_WINDOWS) */
